@@ -29,6 +29,10 @@
 #include <ctype.h>
 #include <stdlib.h>
 #include <string.h>
+#include <iconv.h>
+#include <errno.h>
+
+
 #include "../biosig-dev.h"
 
 #include "structures.h"
@@ -314,6 +318,95 @@ void deallocEN1064(en1064_t en1064) {
 	en1064.Section4.beat = NULL;
 }
 
+
+int decode_scp_text(HDRTYPE *hdr, size_t inbytesleft, char *input, size_t outbytesleft, char *output) {
+
+#if  defined(_ICONV_H) || defined (_LIBICONV_H)
+/*
+	decode_scp_text converts SCP text strings into UTF-8 strings
+	The table of language support code as defined in
+	CEN Standard EN1064:2005+A1:2007, p.30.
+*/
+	uint8_t LanguageSupportCode = (*(struct aecg*)(hdr->aECG)).Section1.Tag14.LANG_SUPP_CODE;
+	iconv_t cd;
+	if ((LanguageSupportCode & 0x01) == 0)
+		cd = iconv_open ("UTF-8", "ASCII");
+
+	else if ((LanguageSupportCode & 0x03) == 1)
+		cd = iconv_open ("UTF-8", "ISO8859-1");
+
+	else if (LanguageSupportCode == 0x03)
+		cd = iconv_open ("UTF-8", "ISO8859-2");
+
+	else if (LanguageSupportCode == 0x0b)
+		cd = iconv_open ("UTF-8", "ISO8859-4");
+
+	else if (LanguageSupportCode == 0x13)
+		cd = iconv_open ("UTF-8", "ISO8859-5");
+
+	else if (LanguageSupportCode == 0x1b)
+		cd = iconv_open ("UTF-8", "ISO8859-6");
+
+	else if (LanguageSupportCode == 0x23)
+		cd = iconv_open ("UTF-8", "ISO8859-7");
+
+	else if (LanguageSupportCode == 0x2b)
+		cd = iconv_open ("UTF-8", "ISO8859-8");
+
+	else if (LanguageSupportCode == 0x33)
+		cd = iconv_open ("UTF-8", "ISO8859-11");
+
+	else if (LanguageSupportCode == 0x3b)
+		cd = iconv_open ("UTF-8", "ISO8859-15");
+
+	else if (LanguageSupportCode == 0x07)
+		cd = iconv_open ("UTF-8", "ISO-10646");
+
+	else if (LanguageSupportCode == 0x0f)	// JIS X 0201-1976 (Japanese) - does not match exactly
+		cd = iconv_open ("UTF-8", "EUC-JISX0213");
+	else if (LanguageSupportCode == 0x17)	// JIS X 0208-1997 (Japanese) - does not match exactly
+		cd = iconv_open ("UTF-8", "EUC-JISX0213");
+	else if (LanguageSupportCode == 0x1f)	// JIS X 0212-1990 (Japanese) - does not match exactly
+		cd = iconv_open ("UTF-8", "EUC-JISX0213");
+
+	else if (LanguageSupportCode == 0x27)
+		cd = iconv_open ("UTF-8", "GB2312");
+
+	else if (LanguageSupportCode == 0x2F)  // KS C5601-1987 (Korean) - does not match exactly
+		cd = iconv_open ("UTF-8", "EUC-KR");
+	else {
+		biosigERROR(hdr, B4C_CHAR_ENCODING_UNSUPPORTED, "SCP character encoding not supported");
+		return -1;
+	}
+
+	if (VERBOSE_LEVEL>7) fprintf(stdout,"%s(%i) decode_scp_text: input=<%s>\n", __FILE__, __LINE__, input);
+
+	iconv(cd, &input, &inbytesleft, &output, &outbytesleft);
+	int errsv = errno;
+
+	if (errsv)
+		biosigERROR(hdr, B4C_CHAR_ENCODING_UNSUPPORTED, "conversion of SCP text failed");
+
+
+	if (VERBOSE_LEVEL>7) fprintf(stdout,"%s(%i) decode_scp_text: output=<%s>\n", __FILE__, __LINE__, output);
+
+	return (iconv_close(cd) || errsv);
+
+#else  // if neither _ICONV_H nor _LIBCONV_H are defined
+
+	if ((LanguageSupportCode & 0x01) != 0) {
+		biosigERROR(hdr, B4C_CHAR_ENCODING_UNSUPPORTED, "SCP character encoding not supported");
+	}
+	else	// ASCII encoding is UTF-8 compatible - no convesion needed
+		strncpy(output, input, min(inbytesleft, outbytesleft+1));
+
+	return(LanguageSupportCode & 0x01);
+
+#endif
+
+}
+
+
 EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 /*
 	this function is a stub or placeholder and need to be defined in order to be useful. 
@@ -370,6 +463,7 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 
 	aECG->Section1.Tag14.VERSION = 0; // acquiring.protocol_revision_number
 	aECG->Section1.Tag15.VERSION = 0; // analyzing.protocol_revision_number
+	aECG->Section1.Tag14.LANG_SUPP_CODE = 0;
 	aECG->FLAG.HUFFMAN   = 0;
 	aECG->FLAG.DIFF      = 0;
 	aECG->FLAG.REF_BEAT  = 0;
@@ -447,7 +541,7 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 	section[0].index  = 6+16;
 	int K;
 	for (K=1; K<_NUM_SECTION; K++) {
-		section[K].ID	  = K;
+		section[K].ID	  = -1;
 		section[K].length = 0;
 		section[K].index  = 0;
 	}
@@ -456,12 +550,46 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 		// this is needed because fields are not always sorted
 		curSect = leu32p(ptr+6+16+K*10);
 		if (curSect < _NUM_SECTION) {
+			if (section[curSect].ID >= 0) {
+				biosigERROR(hdr, B4C_FORMAT_UNSUPPORTED, "SCP Section must not be defined twice");
+				return -1;
+			}
 			section[curSect].ID 	= curSect;
 			section[curSect].length = leu32p(ptr+6+16+K*10+2);
 			section[curSect].index  = leu32p(ptr+6+16+K*10+6)-1;
 		}
 	}
 
+	if (section[1].length) {
+		/**** identify language support code - scan through section 1 for tag 14, byte 17 ****/
+		K = 1;
+		curSect           = section[K].ID;
+		len		  = section[K].length;
+		sectionStart 	  = section[K].index;
+
+		PtrCurSect = ptr+sectionStart;
+		crc 	   = leu16p(PtrCurSect);
+		uint16_t tmpcrc = CRCEvaluate((uint8_t*)(PtrCurSect+2),len-2);
+		uint8_t versionSection  = *(ptr+sectionStart+8);
+		uint8_t versionProtocol = *(ptr+sectionStart+9);
+		// future versions might not need to do this, because language encoding is fixed (i.e. known).
+
+			uint32_t len1;
+			curSectPos = 16;
+			while (curSectPos<=len) {
+				tag = *(PtrCurSect+curSectPos);
+				len1 = leu16p(PtrCurSect+curSectPos+1);
+				curSectPos += 3;
+			if (curSectPos+len1 > len) break;
+				if (tag==14) {
+					aECG->Section1.Tag14.LANG_SUPP_CODE  = *(PtrCurSect+curSectPos+16);	// tag 14, byte 16 (LANG_SUPP_CODE has to be 0x00 => Ascii only, 
+					if (VERBOSE_LEVEL>7) fprintf(stdout,"%s (line %i) Language Support Code is 0x%x\n",__FILE__,__LINE__,aECG->Section1.Tag14.LANG_SUPP_CODE);
+					break;
+				}
+				curSectPos += len1;
+			}
+	}
+
 	for (K=1; K<NSections; K++)	{
 
 	if (VERBOSE_LEVEL>7)
@@ -483,6 +611,8 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 		PtrCurSect = ptr+sectionStart;
 		crc 	   = leu16p(PtrCurSect);
 		uint16_t tmpcrc = CRCEvaluate((uint8_t*)(PtrCurSect+2),len-2); 
+		uint8_t versionSection  = *(ptr+sectionStart+8);
+		uint8_t versionProtocol = *(ptr+sectionStart+9);
 #ifndef ANDROID
 		if ((crc != 0xffff) && (crc != tmpcrc))
 			fprintf(stderr,"Warning SOPEN(SCP-READ): faulty CRC in section %i: crc=%x, %x\n" ,curSect,crc,tmpcrc);
@@ -490,10 +620,6 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 			fprintf(stderr,"Warning SOPEN(SCP-READ): Current Section No does not match field in sections (%i %i)\n",curSect,leu16p(PtrCurSect+2)); 
 		if (len != leu32p(PtrCurSect+4))
 			fprintf(stderr,"Warning SOPEN(SCP-READ): length field in pointer section (%i) does not match length field in sections (%i %i)\n",K,len,leu32p(PtrCurSect+4)); 
-#endif
-		uint8_t versionSection  = *(ptr+sectionStart+8);
-		uint8_t versionProtocol = *(ptr+sectionStart+9);
-#ifndef ANDROID
 		if (versionSection != 13 && versionSection != 20)
 			fprintf(stderr,"Warning SOPEN(SCP-READ): Version of section %i is not 13 or 20 but %i. This is not tested.\n", curSect, versionSection);
 		if (versionProtocol != 13 && versionProtocol != 20)
@@ -524,7 +650,7 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 			while ((curSectPos<=len) && (*(PtrCurSect+curSectPos) < 255)) {
 				tag = *(PtrCurSect+curSectPos);
 				len1 = leu16p(PtrCurSect+curSectPos+1);
-				if (VERBOSE_LEVEL>8)
+				if (VERBOSE_LEVEL > 7)
 					fprintf(stdout,"SCP(r): Section 1 Tag %i Len %i\n",tag,len1); 
 
 				curSectPos += 3;
@@ -535,11 +661,23 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 			break;
 				}
 				if (tag==0) {
-					if (!hdr->FLAG.ANONYMOUS)
-						strncpy(hdr->Patient.Name, (char*)(PtrCurSect+curSectPos),min(len1,MAX_LENGTH_NAME));
+					// convert to UTF8
+					if (!hdr->FLAG.ANONYMOUS) {
+						// strncpy(hdr->Patient.Name, (char*)(PtrCurSect+curSectPos),min(len1,MAX_LENGTH_NAME));
+						 // convert to UTF8
+						decode_scp_text(hdr, len1, PtrCurSect+curSectPos, MAX_LENGTH_NAME, hdr->Patient.Name);
+					}
 				}
 				else if (tag==1) {
-//					hdr->Patient.FirstName = (char*)(PtrCurSect+curSectPos);
+					if (!hdr->FLAG.ANONYMOUS) {
+						size_t len = strlen(hdr->Patient.Name);
+						if (len+3 < MAX_LENGTH_NAME) {
+							strcat(hdr->Patient.Name,", ");
+							len+=2;
+							// convert to UTF8
+							decode_scp_text(hdr, len1, PtrCurSect+curSectPos, MAX_LENGTH_NAME-len+1, hdr->Patient.Name+len);
+						}
+					}
 				}
 				else if (tag==2) {
 					if (len1 > MAX_LENGTH_PID) {
@@ -547,13 +685,23 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 						fprintf(stdout,"Warning SCP(read): length of Patient Id (section1 tag2) exceeds %i>%i\n",len1,MAX_LENGTH_PID); 
 #endif
 					}
-					strncpy(hdr->Patient.Id,(char*)(PtrCurSect+curSectPos),min(len1,MAX_LENGTH_PID));
+					// convert to UTF8
+					decode_scp_text(hdr, len1, PtrCurSect+curSectPos, MAX_LENGTH_PID, hdr->Patient.Id);
 					hdr->Patient.Id[MAX_LENGTH_PID] = 0;
 					
 					if (!strcmp(hdr->Patient.Id,"UNKNOWN"))
 						hdr->Patient.Id[0] = 0;
 				}
 				else if (tag==3) {
+					if (!hdr->FLAG.ANONYMOUS) {
+						size_t len = strlen(hdr->Patient.Name);
+						if (len+2 < MAX_LENGTH_NAME) {
+							strcat(hdr->Patient.Name," ");
+							len+=1;
+							// convert to UTF8
+							decode_scp_text(hdr, len1, PtrCurSect+curSectPos, MAX_LENGTH_NAME-len+1, hdr->Patient.Name+len);
+						}
+					}
 				}
 				else if (tag==4) {
 				}
@@ -581,6 +729,7 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 				else if (tag==9) {
 				}
 				else if (tag==10) {
+					// TODO: convert to UTF8
 				}
 				else if (tag==11) {
 					aECG->systolicBloodPressure  = leu16p(PtrCurSect+curSectPos);
@@ -589,9 +738,12 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 					aECG->diastolicBloodPressure = leu16p(PtrCurSect+curSectPos);
 				}
 				else if (tag==13) {
+					// TODO: convert to UTF8
 					aECG->Diagnosis = (char*)(PtrCurSect+curSectPos);
 				}
 				else if (tag==14) {
+					/* Acquiring Device ID Number */
+					// TODO: convert to UTF8
 #ifndef ANDROID
 					if (len1>85)
 						fprintf(stderr,"Warning SCP(r): length of tag14 %i>40\n",len1);
@@ -649,38 +801,65 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 					tmp += strlen((char*)(PtrCurSect+curSectPos+36+tmp+1)); 
 					aECG->Section1.Tag14.ACQ_DEV_MANUF  = (char*)(PtrCurSect+curSectPos+36+tmp+1);	// tag 14, byte 38 (ACQ_DEV_MANUF has to be "Manufacturer")
 					
+
+					if (aECG->Section1.Tag14.LANG_SUPP_CODE & 0x01) {
+#if _ICONV_H
+						fprintf(stdout, "Warning SCP-ECG: decoding of text strings not ready yet");
+#else
+						biosigERROR(hdr, B4C_CHAR_ENCODING_UNSUPPORTED, "SCP-SCP: Non-ASCII text string language - conversion not supported");
+#endif
+					}
+
 					
 				}
 				else if (tag==15) {
+					/* Analyzing Device ID Number */
+					// TODO: convert to UTF8
 					//memcpy(hdr->aECG->Section1.tag15,PtrCurSect+curSectPos,40);
 					aECG->Section1.Tag15.VERSION     = *(PtrCurSect+curSectPos+14);
 				}
 				else if (tag==16) {
-					// hdr->ID.Hospital = strndup(PtrCurSect+curSectPos+14,len1);
-					hdr->ID.Hospital = malloc(len1);
+					/* Acquiring Institution Description */
+					size_t outlen = len1*2+1;
+					hdr->ID.Hospital = malloc(outlen);
 					if (hdr->ID.Hospital) {
-						hdr->ID.Hospital[len1] = 0;
-						memcpy(hdr->ID.Hospital, PtrCurSect+curSectPos+14, len1);
+						// convert to UTF8
+						decode_scp_text(hdr, len1, PtrCurSect+curSectPos, outlen, hdr->ID.Hospital);
+						hdr->ID.Hospital[outlen] = 0;
 					}
 				}
 				else if (tag==17) {
+					/* Analyzing Institution Description */
+					// TODO: convert to UTF8
 				}
 				else if (tag==18) {
+					/* Acquiring Institution Description */
+					// TODO: convert to UTF8
 				}
 				else if (tag==19) {
+					/* Analyzing Institution Description */
+					// TODO: convert to UTF8
 				}
 				else if (tag==20) {
+					// TODO: convert to UTF8
 					aECG->ReferringPhysician = (char*)(PtrCurSect+curSectPos);
 				}
 				else if (tag==21) {
+					// TODO: convert to UTF8
 					aECG->MedicationDrugs = (char*)(PtrCurSect+curSectPos);
 				}
 				else if (tag==22) {
-					hdr->ID.Technician = (char*)realloc(hdr->ID.Technician,len1+1);
-					memcpy(hdr->ID.Technician,(char*)(PtrCurSect+curSectPos),len1);
-					hdr->ID.Technician[len1] = 0;
+					size_t outlen = len1*2+1;
+					hdr->ID.Technician = malloc(outlen);
+					if (hdr->ID.Technician) {
+						// convert to UTF8
+						decode_scp_text(hdr, len1, PtrCurSect+curSectPos, outlen, hdr->ID.Technician);
+						hdr->ID.Technician[outlen] = 0;
+					}
 				}
 				else if (tag==23) {
+					/* Room Description */
+					// TODO: convert to UTF8
 				}
 				else if (tag==24) {
 					aECG->EmergencyLevel = *(PtrCurSect+curSectPos);
@@ -713,10 +892,16 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 						Notch = 50.0; 	// notch 50Hz
 				}
 				else if (tag==30) {
+					/* Free Text Field */
+					// TODO: convert to UTF8
 				}
 				else if (tag==31) {
+					/* ECG Sequence Number */
+					// TODO: convert to UTF8
 				}
 				else if (tag==32) {
+					/* History Diagnostic Codes */
+					// TODO: convert to UTF8
 					if (PtrCurSect[curSectPos]==0) {
 						unsigned k=1;
 						for (; k < len1; k++) {
@@ -732,8 +917,12 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 					}
 				}
 				else if (tag==33) {
+					/* Electrode Configuration Code */
+					// TODO: convert to UTF8
 				}
 				else if (tag==34) {
+					/* DateTimeZone */
+					// TODO: convert to UTF8
 					int16_t tzmin = lei16p(PtrCurSect+curSectPos);
 					if (tzmin != 0x7fff) {
 						if (abs(tzmin)<=780)
@@ -743,6 +932,10 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 					}
 					//fprintf(stdout,"SOPEN(SCP-READ): tzmin = %i %x \n",tzmin,tzmin);
 				}
+				else if (tag==35) {
+					/* Free Text Medical History */
+					// TODO: convert to UTF8
+				}
 				else {
 				}
 				curSectPos += len1;
@@ -807,14 +1000,12 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 		{
 			hdr->NS = *(PtrCurSect+curSectPos);
 			aECG->FLAG.REF_BEAT = (*(PtrCurSect+curSectPos+1) & 0x01);
-			en1064.FLAG.REF_BEAT = (*(PtrCurSect+curSectPos+1) & 0x01);
 			en1064.Section3.flags = *(PtrCurSect+curSectPos+1);
-			if (en1064.FLAG.REF_BEAT && !section[4].length) {
+			if (aECG->FLAG.REF_BEAT && !section[4].length) {
 #ifndef ANDROID
 				fprintf(stderr,"Warning (SCP): Reference Beat but no Section 4\n");
 #endif
 				aECG->FLAG.REF_BEAT  = 0;
-				en1064.FLAG.REF_BEAT = 0;
 			}
 #ifndef ANDROID
 			if (!(en1064.Section3.flags & 0x04) || ((en1064.Section3.flags>>3) != hdr->NS))
@@ -871,7 +1062,7 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 				en1064.Section4.SPR += en1064.Section4.beat[i].QE-en1064.Section4.beat[i].QB-1;
 			}
 			if (en1064.Section4.len_ms==0) {
-				en1064.FLAG.REF_BEAT = 0;
+				aECG->FLAG.REF_BEAT  = 0;
 			}
 		}
 
@@ -896,7 +1087,7 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 			}
 
 			en1064.Section5.datablock = NULL;
-			if (en1064.FLAG.REF_BEAT) {
+			if (aECG->FLAG.REF_BEAT) {
 				en1064.Section5.datablock = (int32_t*)malloc(4 * hdr->NS * en1064.Section5.Length);
 
 				Ptr2datablock           = (PtrCurSect+curSectPos+6+2*hdr->NS);
@@ -1016,7 +1207,7 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 						data[ix] += 2*data[ix-1] - data[ix-2];
 				}
 #ifndef WITHOUT_SCP_DECODE
-				if (aECG->FLAG.BIMODAL || en1064.FLAG.REF_BEAT) {
+				if (aECG->FLAG.BIMODAL || aECG->FLAG.REF_BEAT) {
 //				if (aECG->FLAG.BIMODAL) {
 //				if (aECG->FLAG.REF_BEAT {
 					/*	this is experimental work
@@ -1048,7 +1239,7 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 					} while (k2 && (k1>0));
 				}
 
-				if (en1064.FLAG.REF_BEAT) {
+				if (aECG->FLAG.REF_BEAT) {
 					/* Add reference beats */
 					// ### FIXME ###
 					for (k1 = 0; k1 < en1064.Section4.N; k1++) {
@@ -1162,6 +1353,7 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 
 		/**** SECTION 8 ****/
 		else if (curSect==8)  {
+			// TODO: convert to UTF8
 #if (BIOSIG_VERSION >= 10500)
 			hdr->SCP.Section8Length = leu32p(PtrCurSect+4)-curSectPos;
 			hdr->SCP.Section8 = PtrCurSect+curSectPos;
@@ -1187,9 +1379,10 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 
 		/**** SECTION 9 ****/
 		else if (curSect==9)  {
+			// TODO: convert to UTF8
 #if (BIOSIG_VERSION >= 10500)
 			hdr->SCP.Section9Length = leu32p(PtrCurSect+4)-curSectPos;
-			hdr->SCP.Section9 = PtrCurSect+curSectPos;
+			hdr->SCP.Section9       = PtrCurSect+curSectPos;
 #else
 			aECG->Section9.StartPtr = (char*)(PtrCurSect+curSectPos);
 			aECG->Section9.Length   = len;
@@ -1206,6 +1399,7 @@ EXTERN_C int sopen_SCP_read(HDRTYPE* hdr) {
 
 		/**** SECTION 11 ****/
 		else if (curSect==11)  {
+			// TODO: convert to UTF8
  			if(len<curSectPos+9) continue; //Something is very wrong
 #if (BIOSIG_VERSION >= 10500)
 			/*
